use std::time::Duration;

use moka::{future::Cache, notification::RemovalCause};
use once_cell::sync::Lazy;
use rocket_prometheus::{prometheus::{IntCounterVec, opts, Error}, PrometheusMetrics};
use std::hash::Hash;

use super::pdf::PdfStream;

static CACHE_EVICTION_COUNT: Lazy<IntCounterVec> = Lazy::new(||
    IntCounterVec::new(opts!(format!("{}_{}", std::env::var("ROCKET_PROMETHEUS_NAMESPACE").unwrap_or("app".to_string()), "cache_eviction"), "Count of cache eviction"), &["cause"])
        .expect("Could not create NAME_COUNTER")
);
static CACHE_HITS: Lazy<IntCounterVec> = Lazy::new(||
        IntCounterVec::new(opts!(format!("{}_{}", std::env::var("ROCKET_PROMETHEUS_NAMESPACE").unwrap_or("app".to_string()), "cache_hists"), "Count of cache eviction"), &[]).expect("Could not create NAME_COUNTER")
);
static CACHE_MISSES: Lazy<IntCounterVec> = Lazy::new(||
        IntCounterVec::new(opts!(format!("{}_{}", std::env::var("ROCKET_PROMETHEUS_NAMESPACE").unwrap_or("app".to_string()), "cache_misses"), "Count of cache eviction"), &[]).expect("Could not create NAME_COUNTER")
);

pub fn cache_prometheus_metrics(metrics: &PrometheusMetrics) -> Result<(), Error> {
    metrics.registry().register(Box::new(CACHE_EVICTION_COUNT.clone()))?;
    metrics.registry().register(Box::new(CACHE_HITS.clone()))?;
    metrics.registry().register(Box::new(CACHE_MISSES.clone()))?;
    Ok(())
}

pub fn build_cache() -> NanoCache<String, PdfStream> {
    let cache = Cache::builder()
        .time_to_live(Duration::from_secs(60 * 60))
        .eviction_listener(|_k, _v, cause| {
            let cause_string = match cause {
                RemovalCause::Expired => "expired",
                RemovalCause::Explicit => "explicit",
                RemovalCause::Replaced => "replaced",
                RemovalCause::Size => "size",
            };

            CACHE_EVICTION_COUNT.with_label_values(&[cause_string]).inc();
        })
        .build();

    NanoCache::new(cache)

}

pub struct NanoCache<K, V> {
    inner: Cache<K, V>,
}

impl<K, V> NanoCache<K, V>
where
    K: Hash + Eq + Send + Sync + 'static,
    V: Clone + Send + Sync + 'static
{
    pub fn new(cache: Cache<K, V>) -> Self {
        Self {
            inner: cache
        }
    }

    pub async fn get(&self, key: &K, force_invalidate: Option<bool>) -> Option<V> {
        if force_invalidate == Some(true) {
            self.inner.invalidate(key).await;
            CACHE_MISSES.with_label_values(&[]).inc();
            return None;
        }
        let result = self.inner.get(key).await;

        match result {
            Some(_) => CACHE_HITS.with_label_values(&[]).inc(),
            None =>  CACHE_MISSES.with_label_values(&[]).inc(),
        }

        result
    }

    pub async fn insert(&self, key: K, value: V) {
        self.inner.insert(key, value).await
    }
}
